import csv
from phonepiece.config import PhonePieceConfig
from phonepiece.arpa import ArpaConverter


class TIMITConverter:

    def __init__(self):

        timit_path = PhonePieceConfig.data_path / 'timit.txt'

        self.timit_map = {}

        with open(timit_path, 'r') as f:
            for line in f:
                fields = line.strip().split()

                if fields[1] != 'nan':
                    self.timit_map[fields[0]] = fields[2]
                else:
                    self.timit_map[fields[0]] = ''

        self.arpa_converter = ArpaConverter()

        self.cache = {}

    def convert(self, str_or_lst):

        if isinstance(str_or_lst, str):
            str_or_lst = str_or_lst.split(' ')

        result = []
        for phone in str_or_lst:
            phone = phone.lower()

            if phone in self.cache:
                result.extend(self.cache[phone])
                continue

            if phone not in self.timit_map:
                PhonePieceConfig.logger.error(f"{phone} is not a TIMIT phone")

            timit_phone = self.timit_map[phone]

            # skip q and silence
            if timit_phone == '' or timit_phone == 'sil':
                continue

            ipa_phone = self.arpa_converter.convert(timit_phone)
            self.cache[phone] = ipa_phone

            result.extend(ipa_phone)

        return result


import unittest
from phonepiece.inventory import read_inventory


class TestInventory(unittest.TestCase):

    def test_all(self):
        inv = read_inventory('eng')

        phones = inv.phone.tolist()
        self.assertTrue(phones[0] == '<blk>')
        self.assertTrue(phones[-1] == '<eos>')

        phonemes = inv.phoneme.tolist()
        self.assertTrue(phonemes[0] == '<blk>')
        self.assertTrue(phonemes[-1] == '<eos>')


import platform
from pathlib import Path
import logging


class PhonePieceConfig:

    root_path = Path(__file__).parent

    # data path
    data_path = root_path / 'data'
    data_str = str(data_path)

    logger = logging.getLogger('phonepiece')


import unittest
from phonepiece.epitran import read_epitran_g2p


class TestEpitran(unittest.TestCase):

    def test_all(self):
        epi = read_epitran_g2p('spa')
        self.assertTrue(epi['b'] == ['b'])


import sys

def print_result(string_lst, out=None):

    res = ""
    for str in string_lst:
        res += '{0: >5}'.format(str)

    out.write(res+'\n')

def fuzzy_match(pattern, text):
    """
    find the best edit distance

    :param string_a: pattern
    :param string_b: text
    :return: matched string part
    """

    string_b = pattern
    string_a = text

    # length of each string
    len_a = len(string_a)
    len_b = len(string_b)

    # dp table
    # its contents is edit distance, dx, dy to the best previous path
    dp = [[(0, -1, -1) for x in range(len_a+1)] for y in range(len_b+1)]

    # initialize first row and first column
    for i in range(len_a+1):
        dp[0][i] = [0, 0, 0]

    for i in range(len_b+1):
        dp[i][0] = [i, -1, 0]

    # dp update
    for i in range(1, len_b+1):
        for j in range(1, len_a+1):
            index_a = j-1
            index_b = i-1
            cost = 0 if string_a[index_a] == string_b[index_b] else 1

            dx = 0
            dy = 0
            mincost = 0

            if dp[i-1][j][0] < dp[i][j-1][0]:
                dx = -1
                dy = 0
                mincost = dp[i-1][j][0] + 1
            else:
                dx = 0
                dy = -1
                mincost = dp[i][j-1][0] + 1

            if dp[i-1][j-1][0] + cost < mincost:
                dx = -1
                dy = -1
                mincost = dp[i-1][j-1][0] + cost

            dp[i][j] = [mincost, dx, dy]


    mincost = 10000000
    best_start_index = 0
    best_end_index = len_b

    for i in range(len_a+1):
        if dp[len_b][i][0] < mincost:
            mincost = dp[len_b][i][0]
            best_end_index = i


    # backward to get the best_start_index
    cx = len_b
    cy = best_end_index

    while cx != 0:
        dx = dp[cx][cy][1]
        dy = dp[cx][cy][2]

        cx += dx
        cy += dy

    best_start_index = cy
    best_end_index -= 1

    return mincost, best_start_index, best_end_index

def edit_distance(string_a, string_b, utt_id='utt_id', verbose=False, out=None):

    # length of each string
    len_a = len(string_a)
    len_b = len(string_b)

    # dp table
    dp = [[0 for x in range(len_a+1)] for y in range(len_b+1)]
    path = [[(0, 0) for x in range(len_a+1)] for y in range(len_b+1)]

    # initialize first row and first column
    for i in range(1, len_a+1):
        dp[0][i] = i
        path[0][i] = (0, i-1)

    for i in range(1, len_b+1):
        dp[i][0] = i
        path[i][0] = (i-1, 0)

    # dp update
    for i in range(1, len_b+1):
        for j in range(1, len_a+1):
            index_a = j-1
            index_b = i-1
            cost = 0.0 if string_a[index_a] == string_b[index_b] else 1.0

            dp[i][j] = dp[i-1][j-1]+cost
            path[i][j] = (i-1,j-1)

            if dp[i][j] > dp[i-1][j]+1:
                dp[i][j] = dp[i-1][j]+1
                path[i][j] = (i-1,j)

            if dp[i][j] > dp[i][j-1]+1:
                dp[i][j] = dp[i][j-1]+1
                path[i][j] = (i, j-1)

    hyp_lst = []
    ref_lst = []
    ops_lst = []

    cur_node = (len_b, len_a)

    add_cnt = 0
    sub_cnt = 0
    del_cnt = 0

    add_lst = []
    del_lst = []
    sub_lst = []

    while(cur_node != (0,0)):
        prev_node = path[cur_node[0]][cur_node[1]]

        # substitution or match
        if prev_node[0]+1 == cur_node[0] and prev_node[1]+1 == cur_node[1]:
            hyp_lst.append(string_b[cur_node[0]-1])
            ref_lst.append(string_a[cur_node[1]-1])

            if string_a[cur_node[1]-1] == string_b[cur_node[0]-1]:
                ops_lst.append('')
            else:
                sub_lst.append((string_a[cur_node[1]-1], string_b[cur_node[0]-1]))
                ops_lst.append('Sub')
                sub_cnt += 1

        # addition
        if prev_node[0] + 1 == cur_node[0] and prev_node[1] == cur_node[1]:
            hyp_lst.append(string_b[cur_node[0]-1])
            ref_lst.append('')
            ops_lst.append('Add')
            add_lst.append(string_b[cur_node[0]-1])
            add_cnt += 1

        # deletion
        if prev_node[0] == cur_node[0] and prev_node[1]+1 == cur_node[1]:
            hyp_lst.append('')
            ref_lst.append(string_a[cur_node[1]-1])
            ops_lst.append('Del')
            del_lst.append(string_a[cur_node[1]-1])
            del_cnt += 1

        cur_node = prev_node


    if verbose:

        if out is None:
            out = sys.stdout
        out.write("-"*80+'\n')
        out.write(f"UTT: {utt_id} - add {add_cnt}, del {del_cnt}, sub {sub_cnt}\n")
        hyp_lst.append("HYP: ")
        ref_lst.append("REF: ")
        ops_lst.append("OPS: ")

        hyp_lst.reverse()
        ref_lst.reverse()
        ops_lst.reverse()

        print_result(ref_lst, out)
        print_result(hyp_lst, out)
        print_result(ops_lst, out)

    return dp[len_b][len_a], add_cnt, del_cnt, sub_cnt, add_lst, del_lst, sub_lst


def wer(string_a, string_b):

    # length of each string
    len_a = len(string_a)
    len_b = len(string_b)

    # dp table
    dp = [[0 for x in range(len_a+1)] for y in range(len_b+1)]

    # initialize first row and first column
    for i in range(len_a+1):
        dp[0][i] = i

    for i in range(len_b+1):
        dp[i][0] = i

    # dp update
    for i in range(1, len_b+1):
        for j in range(1, len_a+1):
            index_a = j-1
            index_b = i-1
            cost = 0.0 if string_a[index_a] == string_b[index_b] else 1.0
            dp[i][j] = min(dp[i-1][j-1]+cost, min(dp[i-1][j]+1, dp[i][j-1]+1))

    return dp[len_b][len_a]



def group_edit_distance(string_b, string_a, grp_dict, grp_err, grp_cnt):
    # length of each string
    len_a = len(string_a)
    len_b = len(string_b)

    # dp table
    dp = [[0 for x in range(len_a + 1)] for y in range(len_b + 1)]
    best_path = [[(0, 0) for x in range(len_a + 1)] for y in range(len_b + 1)]

    # initialize first row and first column
    for i in range(len_a + 1):
        dp[0][i] = i
        best_path[0][i] = (0, i - 1)

    for i in range(len_b + 1):
        dp[i][0] = i
        best_path[i][0] = (i - 1, 0)

    # dp update
    for i in range(1, len_b + 1):
        for j in range(1, len_a + 1):
            index_a = j - 1
            index_b = i - 1
            cost = 0.0 if string_a[index_a] == string_b[index_b] else 1.0
            dp[i][j] = min(dp[i - 1][j - 1] + cost, min(dp[i - 1][j] + 1, dp[i][j - 1] + 1))

            # get best path
            if dp[i - 1][j - 1] + cost == dp[i][j]:
                if cost == 0:
                    best_path[i][j] = (i - 1, j - 1)
                else:
                    best_path[i][j] = (-1, -1)

            elif dp[i][j] + 1 == dp[i - 1][j] + 1:
                best_path[i][j] = (i - 1, j)
            else:
                best_path[i][j] = (i, j - 1)

    # compute errors
    i = len_b
    j = len_a

    while i > 0 or j > 0:
        index_b = i - 1
        if best_path[i][j] == (i - 1, j - 1):
            grp_cnt[grp_dict[string_b[index_b]]] += 1
            i = i-1
            j = j-1
        elif best_path[i][j] == (-1, -1):
            #grp_cnt[grp_dict[string_b[index_b]]] += 1
            #grp_err[grp_dict[string_b[index_b]]] += 1

            i = i-1
            j = j-1
        elif best_path[i][j] == (i - 1, j):
            grp_cnt[grp_dict[string_b[index_b]]] += 1
            grp_err[grp_dict[string_b[index_b]]] += 1

            i, j = best_path[i][j]

        else:

            i, j = best_path[i][j]


def analyze_group_edit_distance(ans_path, hyp_path, grp_path):
    ans = dict()
    hyp = dict()

    grp_set = dict()
    grp_err = dict()
    grp_cnt = dict()

    for line in open(ans_path):
        utt_id, sent = line.split(' ', 1)
        words = sent.split()

        if '<unk>' in words:
            continue

        ans[utt_id] = words


    for line in open(hyp_path):
        utt_id, sent = line.split(' ', 1)
        words = sent.split()

        if '<unk>' in words:
            continue

        hyp[utt_id] = words

    for i, line in enumerate(open(grp_path)):
        phonemes = line.split()

        for phone in phonemes:
            grp_set[phone] = i
            grp_err[i] = 0
            grp_cnt[i] = 0

    for k, v in ans.items():
        if k not in hyp:
            print("Not found ", k, ' in hypothesis')
            continue
        group_edit_distance(ans[k], hyp[k], grp_set, grp_err, grp_cnt)

    for i in range(2):
        print("group ", i, "  wer: ", grp_err[i] * 100.0 / grp_cnt[i])



from phonepiece.config import *
from phonepiece.unit import read_unit, write_unit
from phonepiece.ipa import read_ipa
from phonepiece.iso import normalize_lang_id
from collections import defaultdict
from phonepiece.utils import load_lang_dir

def read_inventory(lang_id_or_path, model_name='latest'):
    """
    read inventory of the lang_id from a prebuilt inventory model or a customized local path

    :param lang_id: a 3 char or 2 char iso id
    :param model_name: model name or a customized path
    :type lang_id_or_path:
    :return:
    :rtype:
    """

    if Path(lang_id_or_path).exists():
        lang_dir = Path(lang_id_or_path)
        lang_id = lang_dir.stem
    else:
        # load or download lang dir
        assert len(lang_id_or_path) == 2 or len(lang_id_or_path) == 3

        lang_id = lang_id_or_path
        lang_dir = load_lang_dir(lang_id, model_name)

    # normalize language id (e.g: 2 char 639-1 -> 3 char 639-3)
    lang_id = normalize_lang_id(lang_id)

    phone_unit = read_unit(lang_dir / 'phone.txt')
    phoneme_unit = read_unit(lang_dir / 'phoneme.txt')
    phone2phoneme = defaultdict(list)
    phoneme2phone = defaultdict(list)

    # use allophone file if allovera supports it
    for line in open(lang_dir / 'allophone.txt', encoding='utf-8'):
        fields = line.strip().split()
        phoneme = fields[0]

        for phone in fields[1:]:
            phone2phoneme[phone].append(phoneme)
            phoneme2phone[phoneme].append(phone)

    # blk and eos would be considered as phone/phonemes as well
    phone2phoneme['<blk>'] = ['<blk>']
    phone2phoneme['<eos>'] = ['<eos>']
    phoneme2phone['<blk>'] = ['<blk>']
    phoneme2phone['<eos>'] = ['<eos>']

    return Inventory(lang_id, model_name, phoneme_unit, phone_unit, phone2phoneme, phoneme2phone)


def write_inventory(inv, inv_path):

    inv_path = Path(inv_path)
    inv_path.mkdir(parents=True, exist_ok=True)

    write_unit(inv.phone, inv_path / 'phone.txt')
    write_unit(inv.phoneme, inv_path / 'phoneme.txt')

    allophone = dict()

    for k,v in inv.phone2phoneme.items():
        for phoneme in v:
            if phoneme not in allophone:
                allophone[phoneme] = []

            allophone[phoneme].append(k)

    w = open(inv_path / 'allophone.txt', 'w')

    for phoneme in inv.phoneme.elems[1:-1]:
        w.write(phoneme+' '+' '.join(allophone[phoneme])+'\n')

    w.close()


def is_inventory_available(lang_id, model_name='latest'):

    # check whether a customized path is used or not
    if (Path(model_name) / lang_id).exists():
        return True
    else:
        lang_dir = PhonePieceConfig.data_path / 'model' / model_name / lang_id
        return lang_dir.exists()


class Inventory:

    def __init__(self, lang_id, model_name, phoneme, phone, phone2phoneme, phoneme2phone):
        self.lang_id = lang_id
        self.phoneme = phoneme
        self.phone = phone
        self.phone2phoneme = phone2phoneme
        self.phoneme2phone = phoneme2phone

        self.ipa = read_ipa()
        self.nearest_mapping = dict()
        self.phone_nearest_mapping = dict()

    def __str__(self):
        return f"<Inventory {self.lang_id} phoneme: {len(self.phoneme)}, phone: {len(self.phone)}>"

    def __repr__(self):
        return self.__str__()

    def remap(self, phonemes_or_phones, broad=True):

        remapped_phones = []

        if broad:
            for phoneme in phonemes_or_phones:
                if phoneme is None or len(phoneme.strip()) == 0:
                    continue

                remapped_phones.append(self.get_nearest_phoneme(phoneme))
        else:
            for phone in phonemes_or_phones:
                if phone is None or len(phone.strip()) == 0:
                    continue
                remapped_phones.append(self.get_nearest_phone(phone))

        return remapped_phones

    def phone_to_phoneme(self, phones):
        result = []
        for phone in phones:
            phonemes = self.phone2phoneme[phone]
            result.append(phonemes[0])

        return result

    def validate(self, phonemes):
        for phoneme in phonemes:
            if phoneme not in self.phoneme:
                print("phoneme ", phoneme, " not in phoneme inventory!")
                return False

        return True

    def get_nearest_phone(self, phone, verbose=False):
        """
        map a random phone (may not exist in the inventory) to the nearest phone in the inventory.
        The decision is based on the articulatory distance

        :param phone: a random phone
        :type phone: str
        :return: nearest phone
        :rtype: str
        """

        # special handling for :
        if phone.endswith('ː') and phone[:-1] in self.phone.unit_to_id:
            self.phone_nearest_mapping[phone] = phone[:-1]
            return phone[:-1]

        if phone in self.phone_nearest_mapping:
            nearest_phone = self.phone_nearest_mapping[phone]

        else:

            target_phones = list(self.phone.unit_to_id.keys())[1:-1]
            nearest_phone = self.ipa.most_similar(phone, target_phones, verbose=verbose)
            self.phone_nearest_mapping[phone] = nearest_phone

        return nearest_phone

    def get_nearest_phoneme(self, phoneme):
        """
        map a random phoneme (may not exist in the inventory) to the nearest phoneme in the inventory.
        The decision is based on the articulatory distance

        :param phoneme: a random phoneme
        :type phoneme: str
        :return: nearest phoneme
        :rtype: str
        """

        # special handling for :
        if phoneme.endswith('ː') and phoneme[:-1] in self.phoneme.unit_to_id:
            self.nearest_mapping[phoneme] = phoneme[:-1]
            return phoneme[:-1]

        if phoneme in self.nearest_mapping:
            nearest_phoneme = self.nearest_mapping[phoneme]

        else:

            target_phonemes = list(self.phoneme.unit_to_id.keys())[1:-1]
            nearest_phoneme = self.ipa.most_similar(phoneme, target_phonemes)
            self.nearest_mapping[phoneme] = nearest_phoneme

        return nearest_phoneme